@@ -41,18 +41,20 @@ typedef RE_UINT32 (*RE_GetPropertyFunc)(RE_UINT32 ch);
 
 #define RE_PROP_ALNUM 0x430001
 #define RE_PROP_ALPHA 0x70001
-#define RE_PROP_ASCII 0x440001
-#define RE_PROP_BLANK 0x450001
+#define RE_PROP_ANY 0x440001
+#define RE_PROP_ASCII 0x450001
+#define RE_PROP_ASSIGNED 0x460001
+#define RE_PROP_BLANK 0x470001
 #define RE_PROP_CNTRL 0xF
 #define RE_PROP_DIGIT 0x9
-#define RE_PROP_GRAPH 0x460001
+#define RE_PROP_GRAPH 0x480001
 #define RE_PROP_LOWER 0x80001
-#define RE_PROP_PRINT 0x470001
+#define RE_PROP_PRINT 0x490001
 #define RE_PROP_PUNCT 0x22
 #define RE_PROP_SPACE 0x190001
 #define RE_PROP_UPPER 0x90001
-#define RE_PROP_WORD 0x480001
-#define RE_PROP_XDIGIT 0x490001
+#define RE_PROP_WORD 0x4A0001
+#define RE_PROP_XDIGIT 0x4B0001
 
 #define RE_BREAK_OTHER 0
 #define RE_BREAK_CR 1
@@ -81,10 +83,10 @@ typedef RE_UINT32 (*RE_GetPropertyFunc)(RE_UINT32 ch);
 #define RE_GBREAK_LV 10
 #define RE_GBREAK_LVT 11
 
-extern char* re_strings[944];
-extern RE_Property re_properties[138];
-extern RE_PropertyValue re_property_values[1625];
-extern RE_GetPropertyFunc re_get_property[75];
+extern char* re_strings[946];
+extern RE_Property re_properties[140];
+extern RE_PropertyValue re_property_values[1641];
+extern RE_GetPropertyFunc re_get_property[77];
 
 RE_UINT32 re_get_general_category(RE_UINT32 ch);
 RE_UINT32 re_get_block(RE_UINT32 ch);
@@ -154,7 +156,9 @@ RE_UINT32 re_get_line_break(RE_UINT32 ch);
 RE_UINT32 re_get_numeric_type(RE_UINT32 ch);
 RE_UINT32 re_get_numeric_value(RE_UINT32 ch);
 RE_UINT32 re_get_alphanumeric(RE_UINT32 ch);
+RE_UINT32 re_get_any(RE_UINT32 ch);
 RE_UINT32 re_get_ascii(RE_UINT32 ch);
+RE_UINT32 re_get_assigned(RE_UINT32 ch);
 RE_UINT32 re_get_blank(RE_UINT32 ch);
 RE_UINT32 re_get_graph(RE_UINT32 ch);
 RE_UINT32 re_get_print(RE_UINT32 ch);
